/*
 * Copyright (C) 2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



package com.android.build.gradle.internal.tasks.multidex

import com.google.common.base.Charsets
import com.google.common.base.Joiner
import com.google.common.collect.Maps
import com.google.common.collect.Sets
import com.google.common.io.Files
import org.gradle.api.DefaultTask
import org.gradle.api.tasks.InputFile
import org.gradle.api.tasks.Optional
import org.gradle.api.tasks.OutputFile
import org.gradle.api.tasks.TaskAction

/**
 * Take a list of classes for the main dex (that was computed before obfuscation),
 * a proguard-generated mapping file and create a new list of classes with the new
 * obfuscated names.
 */
class RetraceMainDexList extends DefaultTask {

    @InputFile
    File mainDexListFile

    @InputFile @Optional
    File mappingFile

    @OutputFile
    File outputFile

    @TaskAction
    void retrace() {

        File mapping = getMappingFile()
        // if there is no mapping file or if it doesn't exist, then we just copy from the main
        // dex list ot the output.
        if (mapping == null) {
            Files.copy(getMainDexListFile(), getOutputFile())
            return
        }

        // load the main class names
        List<String> classes = Files.readLines(getMainDexListFile(), Charsets.UTF_8)

        // load the mapping file and create a dictionary
        List<String> mappingLines = Files.readLines(mapping, Charsets.UTF_8)
        Map<String, String> map = createDict(mappingLines)

        // create the deobfuscated class list. This is a set to detect dups
        // You can have the same class coming from the non-obfuscated list and from
        // the shrinked jar that have different names until we remap them.
        Set<String> deobfuscatedClasses = Sets.newHashSetWithExpectedSize(classes.size())

        for (String clazz : classes) {
            String fullName = map.get(clazz)

            deobfuscatedClasses.add(fullName != null ? fullName : clazz)
        }

        String fileContent =
                Joiner.on(System.getProperty("line.separator")).join(deobfuscatedClasses)
        Files.write(fileContent, getOutputFile(), Charsets.UTF_8)
    }

    static Map<String, String> createDict(List<String> lines) {
        Map<String, String> map = Maps.newHashMap()

        for (String line : lines) {
            if (line.startsWith(" ")) {
                continue
            }

            int pos = line.indexOf(" -> ")
            if (pos == -1) {
                throw new RuntimeException("unable to read mapping file.")
            }

            String fullName = line.substring(0, pos)
            String obfuscatedName = line.substring(pos + 4, line.size() - 1)

            map.put(obfuscatedName.replace(".", "/") + ".class",
                    fullName.replace(".", "/") + ".class")
        }

        return map
    }
}
